"""Includes the Product class, which coordinates a documentation build.
"""

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import os
import logging
from io import BytesIO
import abc

import sh

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class BaseProduct(object):
    """Abstract base class specifying the minimum API for products classes."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def html_dir(self):
        """Directory path of the built HTML product."""
        return

    @abc.abstractproperty
    def doc_dir(self):
        """Directory path of the cloned documentation repository."""
        pass


class Product(BaseProduct):
    """Representation of a documentation product, with logic and state
    for building a documentation project given a
    :class:`ltdmason.manifest.Manifest`. This is intended for Eups-based
    documentation builds from a Jenkins environment.

    Parameters
    ----------
    manifest : :class:`ltdmason.manifest.Manifest`.
        A :class:`ltdmason.manifest.Manifest` instance that defines the
        documentation input.
    build_dir : str
        Directory where documentation will be built.
    """
    # Package directories/files that won't get linked into product doc repo
    # Note that _static/ is handled separately
    package_excludes = ['_static', '_build', '_templates', 'conf.py', '.git']

    def __init__(self, manifest, build_dir):
        super().__init__()
        self.manifest = manifest
        self.build_dir = build_dir

    @property
    def doc_dir(self):
        """Directory path of the cloned documentation repository."""
        return os.path.join(self.build_dir, self.manifest.doc_repo_name)

    @property
    def html_dir(self):
        """Directory path of the built HTML product."""
        return os.path.join(self.doc_dir, '_build', 'html')

    def clone_doc_repo(self):
        """Git clones the Sphinx documentation repository for this build
        product (specified in the :attr:`manifest`) into :attr:`build_dir`.
        """
        # Clone
        clone_out_log = BytesIO()
        clone_err_log = BytesIO()
        git_clone = sh.git.bake(_cwd=self.build_dir)
        git_clone.clone(self.manifest.doc_repo_url, self.doc_dir,
                        _out=clone_out_log,
                        _err=clone_err_log)
        log.debug(clone_out_log.getvalue())
        log.debug(clone_err_log.getvalue())

        # Checkout the appropriate ref
        checkout_out_log = BytesIO()
        checkout_err_log = BytesIO()
        git = sh.git.bake(_cwd=self.doc_dir)
        git.checkout(self.manifest.doc_repo_ref,
                     _out=checkout_out_log,
                     _err=checkout_err_log)
        log.debug(checkout_out_log.getvalue())
        log.debug(checkout_err_log.getvalue())

    def link_package_repos(self):
        """Link the doc/ directories of packages into the ``lsstsw``
        checked-out documunetation repository.

        The ``doc/_static/<pkgname>`` of each package is linked to the
        ``_static/<pkgname>/`` directory in the documentation repository.

        All other content of the package's ``doc/`` directory is linked into
        the ``<pkgname>/`` directory in the root of the documentation repo.
        """
        doc_static_dir = os.path.join(self.doc_dir, '_static')
        if not os.path.exists(doc_static_dir):
            os.makedirs(doc_static_dir)

        for package_name, package_data in self.manifest.packages.items():
            pkg_dir = str(package_data['dir'])

            # validate that the package's doc directory exists; skip if not
            source_doc_dir = os.path.join(pkg_dir, 'doc')
            if not os.path.isdir(source_doc_dir):
                log.debug(
                    'Skipping {0}: no doc/ directory'.format(package_name))
                continue

            # Link _static/<pkgname>
            source_static_dir = os.path.join(source_doc_dir, '_static',
                                             package_name)
            if not os.path.isdir(source_static_dir):
                log.debug(
                    'No _static/ directory found for {0}; skipping'
                    .format(package_name))
            else:
                target = os.path.join(self.doc_dir, '_static', package_name)
                log.debug(
                    'Linking {0} to {1}'.format(source_static_dir, target))
                os.symlink(source_static_dir, target)

            # Link all other entities in the doc/ directory to the package's
            # directory in the documentation repo
            target_doc_dir = os.path.join(self.doc_dir, package_name)
            if not os.path.exists(target_doc_dir):
                os.makedirs(target_doc_dir)
            for entity in os.listdir(source_doc_dir):
                if entity in self.package_excludes:
                    # skips protected dirs like _build, _templates, _static
                    # _static is linked separately, above
                    continue
                src = os.path.join(source_doc_dir, entity)
                target = os.path.join(target_doc_dir, entity)
                log.debug('Linking {0} to {1}'.format(src, target))
                os.symlink(src, target)

    def install_dependencies(self):
        """Install dependencies specific in the doc repo's requirements.txt"""
        if os.path.exists(os.path.join(self.doc_dir, 'requirements.txt')):
            pip_out_log = BytesIO()
            pip_err_log = BytesIO()
            pip = sh.pip.bake(_cwd=self.doc_dir)
            pip.install('-r', 'requirements.txt',
                        _out=pip_out_log,
                        _err=pip_err_log)
            log.debug(pip_out_log.getvalue())
            log.debug(pip_err_log.getvalue())

    def build_sphinx(self):
        """Run the Sphinx build process to produce HTML documentation.

        This method calls ``sphinx-build``, which is installed by Sphinx.
        """
        builder = sh.Command('sphinx-build')
        build_out_log = BytesIO()
        build_err_log = BytesIO()
        builder(self.doc_dir, self.html_dir,
                b='html',  # HTML builder
                a=True,  # build all, without caching
                _out=build_out_log,
                _err=build_err_log)
        log.debug(build_out_log.getvalue())
        log.debug(build_err_log.getvalue())


class TravisProduct(BaseProduct):
    """Representation of a documentation product in the Travis environment.

    This product assuemes the doc project has been cloned and build natively
    by Travis; this product merely reports the build directory to the
    uploader.
    """
    def __init__(self, html_dir):
        super(TravisProduct, self).__init__()
        self._html_dir = html_dir

    @property
    def html_dir(self):
        """Directory path of the built HTML product."""
        return self._html_dir

    @property
    def doc_dir(self):
        # LTD Mason on Travis doesn't need to know about the Sphinx repo
        # since it isn't building sphinx.
        raise NotImplementedError



"""S3 upload/sync utilities."""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import os
import logging
import mimetypes

import boto3

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def upload(bucket_name, path_prefix, source_dir,
           upload_dir_redirect_objects=True,
           surrogate_key=None, acl=None,
           cache_control_max_age=31536000,
           aws_access_key_id=None, aws_secret_access_key=None,
           aws_profile=None):
    """Upload built documentation to S3.

    This function places the contents of the Sphinx HTML build directory
    into the ``/path_prefix/`` directory of an *existing* S3 bucket.
    Existing files on S3 are overwritten; files that no longer exist in the
    ``source_dir`` are deleted from S3.

    S3 credentials are assumed to be stored in a place where boto3 can read
    them, such as :file:`~/.aws/credentials`. `aws_profile_name` allows you
    to select while AWS credential profile you wish to use from the
    :file:`~/.aws/credentials`.
    See http://boto3.readthedocs.org/en/latest/guide/quickstart.html.

    Parameters
    ----------
    bucket_name : str
        Name of the S3 bucket where documentation is uploaded.
    path_prefix : str
        The root directory in the bucket where documentation is stored.
    source_dir : str
        Path of the Sphinx HTML build directory on the local file system.
        The contents of this directory are uploaded into the ``/path_prefix/``
        directory of the S3 bucket.
    upload_dir_redirect_objects : bool, optional
        A feature flag to enable uploading objects to S3 for every directory.
        These objects contain headers ``x-amz-meta-dir-redirect=true`` HTTP
        headers that tell Fastly to issue a 301 redirect from the directory
        object to the '/index.html' in that directory.
    surrogate_key : str, optional
        The surrogate key to insert in the header of all objects
        in the ``x-amz-meta-surrogate-key`` field. This key is used to purge
        builds from the Fastly CDN when Editions change.
        If `None` then no header will be set.
    acl : str, optional
        The pre-canned AWS access control list to apply to this upload.
        Defaults to ``'public-read'``, which allow files to be downloaded
        over HTTP by the public. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
        for an overview of S3's pre-canned ACL lists. Note that ACL settings
        are not validated locally.
    cache_control_max_age : int, optional
        Defaults to 31536000 seconds = 1 year.
    aws_access_key_id : str, optional
        The access key for your AWS account. Also set `aws_secret_access_key`.
    aws_secret_access_key : str, optional
        The secret key for your AWS account.
    aws_profile : str, optional
        Name of AWS profile in :file:`~/.aws/credentials`. Use this instead
        of `aws_access_key_id` and `aws_secret_access_key` for file-based
        credentials.
    """
    log.debug('s3upload.upload({0}, {1}, {2})'.format(
        bucket_name, path_prefix, source_dir))

    session = boto3.session.Session(
        profile_name=aws_profile,
        aws_access_key_id=aws_access_key_id,
        aws_secret_access_key=aws_secret_access_key)
    s3 = session.resource('s3')
    bucket = s3.Bucket(bucket_name)

    metadata = None
    if surrogate_key is not None:
        if metadata is None:
            metadata = {}
        metadata['surrogate-key'] = surrogate_key

    if cache_control_max_age is not None:
        cache_control = 'max-age={0:d}'.format(cache_control_max_age)
    else:
        cache_control = None

    manager = ObjectManager(session, bucket_name, path_prefix)

    for (rootdir, dirnames, filenames) in os.walk(source_dir):
        # name of root directory on S3 bucket
        bucket_root = os.path.relpath(rootdir, start=source_dir)
        if bucket_root in ('.', '/'):
            bucket_root = ''

        # Delete bucket directories that no longer exist in source
        bucket_dirnames = manager.list_dirnames_in_directory(bucket_root)
        for bucket_dirname in bucket_dirnames:
            if bucket_dirname not in dirnames:
                log.debug(('Deleting bucket directory {0}'.format(
                    bucket_dirname)))
                manager.delete_directory(bucket_dirname)

        # Delete files that no longer exist in source
        bucket_filenames = manager.list_filenames_in_directory(bucket_root)
        for bucket_filename in bucket_filenames:
            if bucket_filename not in filenames:
                bucket_filename = os.path.join(bucket_root, bucket_filename)
                log.debug('Deleting bucket file {0}'.format(bucket_filename))
                manager.delete_file(bucket_filename)

        # Upload files in directory
        for filename in filenames:
            local_path = os.path.join(rootdir, filename)
            bucket_path = os.path.join(path_prefix, bucket_root, filename)
            log.debug('Uploading to {0}'.format(bucket_path))
            _upload_file(local_path, bucket_path, bucket,
                         metadata=metadata, acl=acl,
                         cache_control=cache_control)

        # Upload a directory redirect object
        if upload_dir_redirect_objects is True:
            bucket_dir_path = os.path.join(path_prefix, bucket_root)
            bucket_dir_path = bucket_dir_path.rstrip('/')
            if metadata:
                redirect_metadata = dict(metadata)
            else:
                redirect_metadata = {}
            redirect_metadata['dir-redirect'] = 'true'
            _upload_object(bucket_dir_path,
                           content='',
                           bucket=bucket,
                           metadata=redirect_metadata,
                           acl=acl,
                           cache_control=cache_control)


def _upload_file(local_path, bucket_path, bucket,
                 metadata=None, acl=None, cache_control=None):
    """Upload a file to the S3 bucket.

    This function uses the mimetypes module to guess and then set the
    Content-Type and Encoding-Type headers.

    Parameters
    ----------
    local_path : str
        Full path to a file on the local file system.
    bucket_path : str
        Destination path (also known as the key name) of the file in the
        S3 bucket.
    bucket : `boto3` Bucket instance
        S3 bucket.
    metadata : dict, optional
        Header metadata values. These keys will appear in headers as
        ``x-amz-meta-*``.
    acl : str, optional
        A pre-canned access control list. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
    cache_control : str, optional
        The cache-control header value. For example, 'max-age=31536000'.
        ``'
    """
    extra_args = {}
    if acl is not None:
        extra_args['ACL'] = acl
    if metadata is not None:
        extra_args['Metadata'] = metadata
    if cache_control is not None:
        extra_args['CacheControl'] = cache_control

    # guess_type returns None if it cannot detect a type
    content_type, content_encoding = mimetypes.guess_type(local_path,
                                                          strict=False)
    if content_type is not None:
        extra_args['ContentType'] = content_type

    log.debug(str(extra_args))

    obj = bucket.Object(bucket_path)
    # no return status from the upload_file api
    obj.upload_file(local_path, ExtraArgs=extra_args)


def _upload_object(bucket_path, bucket, content='',
                   metadata=None, acl=None, cache_control=None):
    """Upload an arbitrary object to an S3 bucket.

    Parameters
    ----------
    bucket_path : str
        Destination path (also known as the key name) of the file in the
        S3 bucket.
    content : str or bytes
        Object content, optional
    bucket : `boto3` Bucket instance
        S3 bucket.
    metadata : dict, optional
        Header metadata values. These keys will appear in headers as
        ``x-amz-meta-*``.
    acl : str, optional
        A pre-canned access control list. See
        https://docs.aws.amazon.com/AmazonS3/latest/dev/acl-overview.html#canned-acl
    cache_control : str, optional
        The cache-control header value. For example, 'max-age=31536000'.
        ``'
    """
    args = {}
    if metadata is not None:
        args['Metadata'] = metadata
    if acl is not None:
        args['ACL'] = acl
    if cache_control is not None:
        args['CacheControl'] = cache_control
    obj = bucket.Object(bucket_path)
    obj.put(Body=content, **args)


class ObjectManager(object):
    """Manage objects existing in a bucket under a specific ``bucket_root``.

    The ObjectManager maintains information about objects that exist in the
    bucket, and can delete objects that no longer exist in the source.

    Parameters
    ----------
    session : :class:`boto3.session.Session`
        A boto3 session instance provisioned with the correct identities.
    bucket_name : str
        Name of the S3 bucket.
    bucket_root : str
        The version slug is the name root directory in the bucket where
        documentation is stored.
    """
    def __init__(self, session, bucket_name, bucket_root):
        super().__init__()
        s3 = session.resource('s3')
        bucket = s3.Bucket(bucket_name)
        self._session = session
        self._bucket = bucket
        self._bucket_root = bucket_root
        # Strip trailing '/' from bucket_root for comparisons
        if self._bucket_root.endswith('/'):
            self._bucket_root = self._bucket_root.rstrip('/')

    def list_filenames_in_directory(self, dirname):
        """List all file-type object names that exist at the root of this
        bucket directory.

        Parameters
        ----------
        dirname : str
            Directory name in the bucket relative to ``bucket_root/``.

        Returns
        -------
        filenames : list
            List of file names (`str`), relative to ``bucket_root/``, that
            exist at the root of ``dirname``.
        """
        prefix = self._create_prefix(dirname)
        filenames = []
        for obj in self._bucket.objects.filter(Prefix=prefix):
            if obj.key.endswith('/'):
                continue
            obj_dirname = os.path.dirname(obj.key)
            if obj_dirname == prefix:
                # object is at root of directory
                filenames.append(os.path.relpath(obj.key,
                                                 start=prefix))
        return filenames

    def list_dirnames_in_directory(self, dirname):
        """List all names of directories that exist at the root of this
        bucket directory.

        Note that *directories* don't exist in S3; rather directories are
        inferred from path names.

        Parameters
        ----------
        dirname : str
            Directory name in the bucket relative to ``bucket_root``.

        Returns
        -------
        dirnames : list
            List of directory names (`str`), relative to ``bucket_root/``,
            that exist at the root of ``dirname``.
        """
        prefix = self._create_prefix(dirname)
        dirnames = []
        for obj in self._bucket.objects.filter(Prefix=prefix):
            dirname = os.path.dirname(obj.key)
            # if the object is a directory redirect, make it look like a dir
            if dirname == '':
                dirname = obj.key + '/'
            rel_dirname = os.path.relpath(dirname, start=prefix)
            dir_parts = rel_dirname.split('/')
            if len(dir_parts) == 1:
                dirnames.append(dir_parts[0])
        dirnames = list(set(dirnames))
        if '.' in dirnames:
            dirnames.remove('.')
        return dirnames

    def _create_prefix(self, dirname):
        if dirname in ('.', '/'):
            dirname = ''
        # Strips trailing slash from dir prefix for comparisons
        # os.dirname() returns directory names without a trailing /
        prefix = os.path.join(self._bucket_root, dirname)
        if prefix.endswith('/'):
            prefix = prefix.rstrip('/')
        return prefix

    def delete_file(self, filename):
        """Delete a file from the bucket.

        Parameters
        ----------
        filename : str
            Name of the file, relative to ``bucket_root/``.
        """
        key = os.path.join(self._bucket_root, filename)
        objects = list(self._bucket.objects.filter(Prefix=key))
        for obj in objects:
            obj.delete()

    def delete_directory(self, dirname):
        """Delete a directory (and contents) from the bucket.

        Parameters
        ----------
        dirname : str
            Name of the directory, relative to ``bucket_root/``.
        """
        key = os.path.join(self._bucket_root, dirname)
        if not key.endswith('/'):
            key += '/'

        delete_keys = {'Objects': []}
        key_objects = [{'Key': obj.key}
                       for obj in self._bucket.objects.filter(Prefix=key)]
        assert len(key_objects) > 0
        delete_keys['Objects'] = key_objects
        # based on http://stackoverflow.com/a/34888103
        s3 = self._session.resource('s3')
        r = s3.meta.client.delete_objects(Bucket=self._bucket.name,
                                          Delete=delete_keys)
        log.debug(r)
        if 'Errors' in r:
            raise S3Error('S3 could not delete {0}'.format(key))


class S3Error(Exception):
    """General errors in S3 API usage."""
    pass



"""Command line interface for ltd-mason specialized for Travis CI-based builds.
"""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import sys
import os
import logging
import argparse
import textwrap

from .manifest import TravisManifest
from .product import TravisProduct
from .uploader import upload


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def run():
    """Entrypoint for ltd-mason-travis command."""
    args = parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    # Only build when the env var 'LTD_MASON_BUILD' is set to True
    # This is useful for Travis build matrices; we want to specify
    # exactly which test run to run Mason on.
    mason_build_flag = os.getenv('LTD_MASON_BUILD')
    if mason_build_flag is None:
        log.info('ltd-mason-travis skipping build')
        sys.exit(0)
    elif mason_build_flag.lower() != 'true':
        log.info('ltd-mason-travis skipping build')
        sys.exit(0)

    # Verify this is a branch build on Travis; not a PR build.
    travis_pr_flag = os.getenv('TRAVIS_PULL_REQUEST')
    if travis_pr_flag is None:
        raise RuntimeError(
            'TRAVIS_PULL_REQUEST environment variable not found')
    if travis_pr_flag.lower() != 'false':
        # Silently exit so as not to disturb a Travis CI of code tests
        log.info('ltd-mason-travis skipping PR build')
        sys.exit(0)

    manifest = TravisManifest()
    product = TravisProduct(os.path.abspath(os.path.expandvars(args.html_dir)))

    if not args.no_upload:
        upload(manifest, product)


def parse_args():
    """Create an ``argparse.ArgumentParser`` instance that define's ltd-mason's
    command line interface.
    """
    parser = argparse.ArgumentParser(
        prog='ltd-mason-travis',
        description=textwrap.dedent("""
            Build Sphinx product documentation for LSST the Docs.

            ltd-mason-travis is intended to be run from a Travis CI
            environment. Travis should build the Sphinx document. Tell Mason
            where the built HTML is located with the --html-dir argument.

            ltd-mason-travis uses environment variables for configuration:

            LTD_MASON_BUILD
               Set to `true` to declare a build should be run with Mason.
               In a Travis CI build matrix, only one build should run Mason.

            LTD_MASON_PRODUCT
               Name of the LSST the Docs product.

            LTD_MASON_AWS_ID
               AWS access key ID.

            LTD_MASON_AWS_SECRET
               AWS secret access key.

            Note that the AWS credentials specified here must have permission
            to read and write into the S3 buckets managed by the LTD Keeper
            server.

            LTD_KEEPER_URL
               URL of LTD Keeper instance.

            LTD_KEEPER_USER
               Username for LTD Keeper instance.

            LTD_KEEPER_PASSWORD
               Password for LTD Keeper instance.

            ltd-mason-travis also assumes that the following environment
            variables are available in a Travis environment:

            TRAVIS_PULL_REQUEST
               If true, then ltd-mason-travis will exit. Only branch-based
               builds are supported.

            TRAVIS_REPO_SLUG
               To identify the source Git repository (assuming it resides on
               GitHub.

            TRAVIS_BRANCH
               To identify the branch that is being build to associate this
               build with an Edition on LSST the Docs.
            """),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='See https://github.com/lsst-sqre/ltd-mason for more info.')
    parser.add_argument(
        '--html-dir',
        dest='html_dir',
        default='',
        help='Directory of the built HTML site')
    parser.add_argument(
        '--no-upload',
        dest='no_upload',
        default=False,
        action='store_true',
        help='Skip the upload to S3 and ltd-keeper')
    parser.add_argument(
        '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='Full logging of debug messages')
    return parser.parse_args()



"""Upload built documentation to S3 and coordinate with ltd-keeper."""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import os
import logging

import requests

# weird import helps with mocking
from .s3upload import upload as s3upload_upload


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def upload(manifest, product):
    aws_credentials = read_aws_credentials()
    keeper_credentials = read_keeper_credentials()
    keeper_token = get_keeper_token(
        keeper_credentials['keeper_url'],
        keeper_credentials['keeper_username'],
        keeper_credentials['keeper_password'])
    upload_via_keeper(manifest, product,
                      keeper_url=keeper_credentials['keeper_url'],
                      keeper_token=keeper_token,
                      aws_credentials=aws_credentials)


def read_aws_credentials():
    keys = (('aws_profile', 'LTD_MASON_AWS_PROFILE'),
            ('aws_access_key_id', 'LTD_MASON_AWS_ID'),
            ('aws_secret_access_key', 'LTD_MASON_AWS_SECRET'))
    c = {k: os.getenv(envvar) for (k, envvar) in keys}
    if c['aws_access_key_id'] is not None \
            and c['aws_secret_access_key'] is not None:
        del c['aws_profile']
        log.debug('Using $LTD_MASON_AWS_ID and $LTD_MASON_AWS_SECRET')
    else:
        del c['aws_access_key_id']
        del c['aws_secret_access_key']
        if c['aws_profile'] is None:
            log.debug('Assuming default AWS credential setup')
            del c['aws_profile']
        else:
            log.debug('Using $LTD_MASON_AWS_PROFILE')

    return c


def read_keeper_credentials():
    keys = (('keeper_url', 'LTD_KEEPER_URL'),
            ('keeper_username', 'LTD_KEEPER_USER'),
            ('keeper_password', 'LTD_KEEPER_PASSWORD'))
    c = {k: os.getenv(envvar) for (k, envvar) in keys}
    for (k, envvar) in keys:
        if c[k] is None:
            raise RuntimeError('Please set {0}'.format(envvar))
    return c


def get_keeper_token(base_url, username, password):
    """Get a temporary auth token from ltd-keeper."""
    token_endpoint = base_url + '/token'
    r = requests.get(token_endpoint, auth=(username, password))
    if r.status_code != 200:
        raise RuntimeError('Could not authenticate to {0}: error {1:d}\n{2}'.
                           format(base_url, r.status_code, r.json()))
    return r.json()['token']


def upload_via_keeper(manifest, product,
                      keeper_url, keeper_token,
                      aws_credentials=None):
    """Upload built documentation to S3 via ltd-keeper.

    This runs a three-step pipeline:

    1. Register the build on LTD Keeper with ``POST /products/<slug>/``
    2. Upload documentation files to an appropriate location (according to
       LTD Keeper) in an S3 bucket.
    3. Confirm to LTD Keeper that the documentation upload is complete.

    Parameters
    ----------
    manifest : :class:`ltdmason.manifest.Manifest`
        The manifest for the documentation build.
    product : :class:`ltdmason.product.Product`
        The :class:`~ltdmason.product.Product` that built the documentation.
    keeper_url : str
        URL of the ltd-keeper HTTP API service.
    keeper_token : str
        Authorization token for the ltd-keeper instance.
    aws_credentials : dict, optional
        A `dict` specifying an AWS credential configuration. There are three
        variants:

        1. Leave unset to rely on defaults in :file:`~/.aws/credentials`::

        2. Runtime-specified credentials (perhaps through ltd-mason environment
           variables)::

              {'aws_access_key_id': '...',
               'aws_secret_access_key': '...'}

        3. Runtime-specific name of a profile of credentials stored in
           :file:`~/.aws/credentials`::

              {'aws_profile': '...'}

        See http://boto3.readthedocs.org/en/latest/guide/configuration.html
        for information on :file:`~/.aws/credentials`.

    Raises
    ------
    KeeperError
       Any anomaly with LTD Keeper interaction.
    """
    # Register the documentation build for this product
    build_resource = _register_build(manifest, keeper_url, keeper_token)

    # Upload documentation site to S3
    if aws_credentials is None:
        # Fall back to using default AWS credentials the user might have set
        aws_credentials = {}
    s3upload_upload(build_resource['bucket_name'],
                    build_resource['bucket_root_dir'],
                    product.html_dir,
                    surrogate_key=build_resource['surrogate_key'],
                    acl=None,
                    cache_control_max_age=31536000,
                    **aws_credentials)
    log.debug('Upload complete: {0}:{1}'.format(
        build_resource['bucket_name'], build_resource['bucket_root_dir']))

    # Confirm upload to ltd-keeper
    _confirm_upload(build_resource['self_url'], keeper_token)


def _register_build(manifest, keeper_url, keeper_token):
    """Register this documentation build with LTD Keeper

    This registration step tells ltd-mason where to upload the documentation
    files (bucket and directory).

    Raises
    ------
    KeeperError
       Any anomaly with LTD Keeper interaction.
    """
    data = {'git_refs': manifest.refs}
    if manifest.build_id is not None:
        data['slug'] = manifest.build_id
    if manifest.requester_github_handle is not None:
        data['github_requester'] = manifest.requester_github_handle

    r = requests.post(
        keeper_url + '/products/{p}/builds/'.format(
            p=manifest.product_name),
        auth=(keeper_token, ''),
        json=data)

    if r.status_code != 201:
        raise KeeperError(r.json())
    build_info = r.json()
    log.debug(r.json())
    return build_info


def _confirm_upload(build_url, keeper_token):
    """Patch the build on LTD Keeper to say that the upload is successful.

    Raises
    ------
    KeeperError
       Any anomaly with LTD Keeper interaction.
    """
    r = requests.patch(build_url,
                       auth=(keeper_token, ''),
                       json={'uploaded': True})
    if r.status_code != 200:
        raise KeeperError(r)
    log.debug(r.json())


class KeeperError(Exception):
    """Error using the LTD Keeper API."""
    pass



"""Command line interface to add directory redirect objects.
"""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA


import argparse
import os
import textwrap
import logging

import boto3

from .s3upload import _upload_object


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def run():
    """Entrypoint for ltd-mason-make-redirects."""
    args = parse_args()

    logging.basicConfig(level=logging.INFO)

    # create a bucket object
    session = boto3.session.Session(
        aws_access_key_id=args.aws_id,
        aws_secret_access_key=args.aws_secret)
    s3 = session.resource('s3')
    bucket = s3.Bucket(args.bucket)

    directories = []
    for obj in bucket.objects.filter(Prefix=args.base_dir):
        dirname = os.path.dirname(obj.key)
        if dirname:
            directories.append(dirname)
    directories = set(directories)
    for dirname in directories:
        if dirname.endswith('/v') or dirname.endswith('/builds'):
            # Skip {product}/v or {product}/builds directories
            continue
        # create a directory redirect object
        redirect_metadata = {'dir-redirect': 'true'}
        cache_control = 'max-age={0}'.format(31536000)
        logging.info('Making redirect object at {0}'.format(dirname))
        if not args.dry_run:
            _upload_object(dirname,
                           content='',
                           bucket=bucket,
                           metadata=redirect_metadata,
                           acl='public-read',
                           cache_control=cache_control)


def parse_args():
    """Create an ``argparse.ArgumentParser`` instance that defines the
    command line interface for ltd-mason-make-directs.
    """
    parser = argparse.ArgumentParser(
        prog='ltd-mason-make-redirects',
        description=textwrap.dedent("""Bulk-add directory courtesy redirect
objects to an existing LSST the Docs bucket.

These redirect objects are named after directories (without a trailing slash)
and contain an ``x-amx-meta-dir-redirect=true`` HTTP header. The Fastly VCL
code detects when these objects are being requested (e.g. example.com/dir)
and issues a 301 redirect to example.com/dir/index.html).

This script should only be run once to add redirects to existing
builds. The regular ltd-mason and ltd-keeper workflows will maintain redirects
subsequently.
            """),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='See https://github.com/lsst-sqre/ltd-mason for more info.')
    parser.add_argument(
        '--bucket',
        help='LSST the Docs S3 bucket',
        required=True)
    parser.add_argument(
        '--base-dir',
        help='Directory to make redirects in (defaults to all directories)',
        default='')
    parser.add_argument(
        '--aws-id',
        help='AWS access key ID',
        required=True)
    parser.add_argument(
        '--aws-secret',
        help='AWS secret access key',
        required=True)
    parser.add_argument(
        '--dry-run',
        help='Dry-run, prevents objects from being uploaded',
        action='store_true',
        default=False)
    return parser.parse_args()



"""Command line interface for ltd-mason that runs a doc build."""
from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import os
import argparse
import textwrap
import sys
from io import open
import tempfile
import shutil
import logging

from .manifest import Manifest
from .product import Product
from .uploader import upload


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def run_ltd_mason():
    """Command line app: ``ltd-mason``.

    This app is intended to be called from ``buildlsstsw.sh`` and is
    responsible for running the full documentation build and upload
    process.
    """
    args, unknown_args = parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    if args.manifest_path is None:
        # Read manifest from stdin
        manifest_data = sys.stdin.read()
    else:
        # Read manifest from file
        with open(args.manifest_path, mode='r', encoding='utf-8') as f:
            manifest_data = f.read()
    manifest = Manifest(manifest_data)

    if args.build_dir is None:
        # Use a temporary directory by default
        build_dir = tempfile.mkdtemp()
    else:
        # Use a debug directory
        build_dir = os.path.abspath(args.build_dir)
        assert build_dir is not os.getcwd(), "--build-dir can't be CWD"
        if os.path.exists(build_dir):
            shutil.rmtree(build_dir)
        os.makedirs(build_dir)

    product = Product(manifest, build_dir)
    product.clone_doc_repo()
    product.link_package_repos()
    product.install_dependencies()
    product.build_sphinx()

    if not args.no_upload:
        upload(manifest, product)

    if args.build_dir is None:
        shutil.rmtree(build_dir)


def parse_args():
    """Create an ``argparse.ArgumentParser`` instance that define's ltd-mason's
    command line interface.
    """
    parser = argparse.ArgumentParser(
        prog='ltd-mason',
        description=textwrap.dedent("""
            Build Sphinx product documentation for LSST the Docs.

            ltd-mason is driven by a YAML-formatted manifest file. The manifest
            can either be provided as a file using the `--manifest` argument,

            ltd-mason --manifest manifest.yaml

            or piped in,

            cat manifest.yaml | ltd-mason

            ltd-mason's use of Amazon S3 and LTD Keeper are configured with
            environment variables:

            LTD_MASON_AWS_ID
               AWS access key ID.

            LTD_MASON_AWS_SECRET
               AWS secret access key.

            LTD_MASON_AWS_PROFILE
               This variable can be set as an alternative ``LTD_MASON_AWS_ID``
               and ``LTD_MASON_AWS_SECRET``. ``LTD_MASON_AWS_PROFILE`` is the
               name of a profile in `~/.aws/credentials` that contains your
               secret key and ID. See the `boto3 configuration guide
               <http://bit.ly/1WuF7rY>`_ for more information.

            Note that the AWS credentials specified here must have permission
            to read and write into the S3 buckets managed by the LTD Keeper
            server.

            LTD_KEEPER_URL
               URL of LTD Keeper instance.

            LTD_KEEPER_USER
               Username for LTD Keeper instance.

            LTD_KEEPER_PASSWORD
               Password for LTD Keeper instance.
            """),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog='See https://github.com/lsst-sqre/ltd-mason for more info.')
    parser.add_argument(
        '--manifest',
        dest='manifest_path',
        default=None,
        help='Path to YAML manifest file that defines the doc build.')
    parser.add_argument(
        '--no-upload',
        dest='no_upload',
        default=False,
        action='store_true',
        help='Skip the upload to S3 and ltd-keeper; only build the docs')
    parser.add_argument(
        '--build-dir',
        default=None,
        dest='build_dir',
        help='Directory to use for building the documentation. By default a '
             'temporary directory is created an deleted. This manually-set '
             'directory is not deleted to aid debugging. Beware that any '
             'existing content in that directory will be deleted.')
    parser.add_argument(
        '--verbose',
        dest='verbose',
        default=False,
        action='store_true',
        help='Full logging of debug messages')
    args, unknown_args = parser.parse_known_args()
    return args, unknown_args



"""Interface for stack manifest YAML files written by buildlsstsw.sh

The purpose of these YAML files is to isolate ltd-mason from Eups/Scons
builds of the software itself and to merely tell ltd-mason where the built
software can be found, and metadata about the versioning of this Stack.
"""

from __future__ import (division, absolute_import, print_function,
                        unicode_literals)
from builtins import *  # NOQA
from future.standard_library import install_aliases
install_aliases()  # NOQA

import abc
from urllib.parse import urlparse, urlunparse
import os

import jsonschema
import ruamel.yaml
import pkg_resources


class BaseManifest(object):
    """Abstract base class defining the API for a Manifest.

    Manifests specify the input parameters for a documentation build.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractproperty
    def doc_repo_url(self):
        """Git URL for the product's Git documentation repository."""
        return

    @abc.abstractproperty
    def doc_repo_ref(self):
        """Git ref (branch, commit, tag) for the product's Git documentation
        repository (:class:`str`).
        """
        return

    @property
    def doc_repo_name(self):
        """Name of the product's Git documentation repository (:class:`str`).

        For example, a doc repository at
        ``'https://github.com/lsst-sqre/pipelines_doc.git'`` is named
        ``'pipelines_doc'``.
        """
        parts = urlparse(self.doc_repo_url)
        return os.path.splitext(parts.path)[0].split('/')[-1]

    @abc.abstractproperty
    def product_name(self):
        """Name of the documentation product for LTD Keeper."""
        return

    @abc.abstractproperty
    def build_id(self):
        """Build identifier (`str`)."""
        return

    @abc.abstractproperty
    def requester_github_handle(self):
        """GitHub username handle of person who triggered the build. `None`
        if not available.
        """
        return

    @abc.abstractproperty
    def refs(self):
        """`list` of Git refs that define the overal version set of the
        products.
        """
        return

    @abc.abstractproperty
    def packages(self):
        """Dictionary of package names as keys and package data as values.

        Package data is a dict with keys:

        - ``'dir'``: directory where the package was installed by lsstsw. This
          is ensured to be an absolute URL, transforming any relative paths
          in the Manifest, assuming they are relative to the **current
          working directory.**
        - ``'url'``: Git repository URL.
        - ``'ref'``: Git reference for package (branch, commit, tag).
        """
        return


class Manifest(BaseManifest):
    """Representation of a YAML-encoded manifest for an LSST stack product.

    Parameters
    ----------
    yaml_data : str
        Stack manifest, encoded as YAML text.

    Attributes
    ----------
    data : `ruamel.yaml` object
        Manifest dataset loaded by :mod:`ruamel.yamel`.
    """
    def __init__(self, f):
        super().__init__()
        data = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)
        Manifest.validate(data)
        self.data = data

    @property
    def yaml(self):
        """YAML representation of the manifest (:class:`str`)."""
        return ruamel.yaml.dump(self.data, Dumper=ruamel.yaml.RoundTripDumper)

    @property
    def doc_repo_url(self):
        """Git URL for the product's Git documentation repository."""
        return self.data['doc_repo']['url']

    @property
    def doc_repo_ref(self):
        """Git ref (branch, commit, tag) for the product's Git documentation
        repository (:class:`str`).
        """
        return self.data['doc_repo']['ref']

    @property
    def product_name(self):
        """Name of the documentation product."""
        return self.data['product_name']

    @property
    def build_id(self):
        """Build identifier (`str`)."""
        return self.data['build_id']

    @property
    def requester_github_handle(self):
        """GitHub username handle of person who triggered the build. `None`
        if not available.
        """
        if 'requester_github_handle' in self.data:
            return self.data['requester_github_handle']
        else:
            return None

    @property
    def refs(self):
        """`list` of Git refs that define the overal version set of the
        products.
        """
        return self.data['refs']

    @property
    def packages(self):
        """Dictionary of package names as keys and package data as values.

        Package data is a dict with keys:

        - ``'dir'``: directory where the package was installed by lsstsw. This
          is ensured to be an absolute URL, transforming any relative paths
          in the Manifest, assuming they are relative to the **current
          working directory.**
        - ``'url'``: Git repository URL.
        - ``'ref'``: Git reference for package (branch, commit, tag).
        """
        data = {}
        for pkg_name, pkg_data in self.data['packages'].items():
            pkg_data = dict(pkg_data)
            pkg_data['dir'] = os.path.abspath(pkg_data['dir'])
            data[pkg_name] = pkg_data
        return data

    @classmethod
    def validate(self, data):
        """Validate the schema of a parsed YAML manifest."""
        schema = load_manifest_schema()
        jsonschema.validate(data, schema)


def load_manifest_schema():
    resource_args = (__name__, '../manifest_schema.yaml')
    assert pkg_resources.resource_exists(*resource_args)
    yaml_data = pkg_resources.resource_string(*resource_args)
    return ruamel.yaml.load(yaml_data)


class TravisManifest(BaseManifest):
    """Manifest for Travis CI based single doc repo builds.

    Unlike the original :class:`Manifest` that was driven by YAML, the
    :class:`TravisManifest` is driven by environment variables available in
    a Travis CI environment.
    """
    def __init__(self):
        super(TravisManifest, self).__init__()

    @property
    def doc_repo_url(self):
        """Git URL for the product's Git documentation repository derived
        from ``$TRAVIS_REPO_SLUG`` and assumes the repo is hosted on GitHub.
        """
        slug = os.getenv('TRAVIS_REPO_SLUG')
        if slug is None:
            raise RuntimeError('Environment variable TRAVIS_REPO_SLUG not set')
        parts = ('https', 'github.com', slug + '.git', '', '', '')
        url = urlunparse(parts)
        return url

    @property
    def doc_repo_ref(self):
        """Git ref (branch name) for the product's Git documentation
        repository (:class:`str`) derived from ``$TRAVIS_BRANCH``.
        """
        branch = os.getenv('TRAVIS_BRANCH')
        if branch is None:
            raise RuntimeError('Environment variable TRAVIS_BRANCH not set')
        return branch

    @property
    def product_name(self):
        """Name of the documentation product for LTD Keeper derived from
        LTD_MASON_PRODUCT environment variable.
        """
        name = os.getenv('LTD_MASON_PRODUCT')
        if name is None:
            message = 'Environment variable LTD_MASON_PRODUCT not set'
            raise RuntimeError(message)
        return name

    @property
    def build_id(self):
        """Build ID is set to `None` to allow LTD Keeper to set an ID."""
        return None

    @property
    def requester_github_handle(self):
        """The GitHub user triggering a build: this is not available on Travis.
        Set to `None`.
        """
        return None

    @property
    def refs(self):
        """`list` of Git refs that define the overal version set of the
        products. On travis this is a one-item list with the branch name.
        """
        return [self.doc_repo_ref]

    @property
    def packages(self):
        """Not applicable for Travis builds. Set to an empty list."""
        return dict()



from setuptools import setup, find_packages
import os
from io import open


packagename = 'ltd-mason'
description = 'LSST the Docs build tool'
author = 'Jonathan Sick'
author_email = 'jsick@lsst.org'
license = 'MIT'
url = 'https://github.com/lsst-sqre/ltd-mason'
version = '0.2.3'


def read(filename):
    full_filename = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        filename)
    return open(full_filename, mode='r', encoding='utf-8').read()


long_description = read('README.rst')


setup(
    name=packagename,
    version=version,
    description=description,
    long_description=long_description,
    url=url,
    author=author,
    author_email=author_email,
    license=license,
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
    keywords='lsst',
    packages=find_packages(exclude=['docs', 'tests*', 'data']),
    install_requires=['future>=0.15',
                      'ruamel.yaml>=0.10',
                      'sh>=1.11',
                      'boto3>=1.2',
                      'jsonschema>=2.5',
                      'sphinx>=1.4',
                      'requests>=2.9'],
    # package_data={},
    entry_points={
        'console_scripts': [
            'ltd-mason = ltdmason.cli:run_ltd_mason',
            'ltd-mason-travis = ltdmason.traviscli:run',
            'ltd-mason-make-redirects = ltdmason.redirectdircli:run',
        ]
    }
)
